use crate::ProtocolError;
use std::io::Read;
use uuid::Uuid;

/// Packet data reader trait
pub trait DataReader {
    /// Read bytes
    fn read_bytes(&mut self, size: usize) -> Result<Vec<u8>, ProtocolError>;

    /// Read byte
    fn read_byte(&mut self) -> Result<u8, ProtocolError> {
        Ok(self.read_bytes(1)?[0])
    }
    /// Read String
    fn read_string(&mut self) -> Result<String, ProtocolError> {
        let size = self.read_varint()? as usize;
        match String::from_utf8(self.read_bytes(size)?) {
            Ok(i) => Ok(i),
            Err(_) => Err(ProtocolError::StringParseError),
        }
    }
    /// Read Signed byte as i8
    fn read_signed_byte(&mut self) -> Result<i8, ProtocolError> {
        Ok(self.read_byte()? as i8)
    }
    /// Read Unsigned Short as u16
    fn read_unsigned_short(&mut self) -> Result<u16, ProtocolError> {
        self.read_bytes(2)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| u16::from_be_bytes(o))
    }
    /// Read Boolean
    fn read_boolean(&mut self) -> Result<bool, ProtocolError> {
        self.read_byte().map(|o| o == 0x01)
    }
    /// Read Short as i16
    fn read_short(&mut self) -> Result<i16, ProtocolError> {
        self.read_bytes(2)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| i16::from_be_bytes(o))
    }
    /// Read Long as i64
    fn read_long(&mut self) -> Result<i64, ProtocolError> {
        self.read_bytes(8)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| i64::from_be_bytes(o))
    }
    /// Read Float as f32
    fn read_float(&mut self) -> Result<f32, ProtocolError> {
        self.read_bytes(4)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| f32::from_be_bytes(o))
    }
    /// Read Double as f64
    fn read_double(&mut self) -> Result<f64, ProtocolError> {
        self.read_bytes(8)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| f64::from_be_bytes(o))
    }
    /// Read Int as i32
    fn read_int(&mut self) -> Result<i32, ProtocolError> {
        self.read_bytes(4)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| i32::from_be_bytes(o))
    }
    /// Read UUID
    fn read_uuid(&mut self) -> Result<Uuid, ProtocolError> {
        self.read_bytes(16)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| Uuid::from_bytes(o))
    }

    /// Read VarInt as i32 with size in bytes (varint, size)
    fn read_varint_size(&mut self) -> Result<(i32, usize), ProtocolError> {
        let mut value: u32 = 0;
        let mut position: u32 = 0;
        let mut size = 0;

        loop {
            let byte = self.read_byte()?;
            value |= ((byte & 0x7F) as u32) << position;

            size += 1;

            if (byte & 0x80) == 0 {
                break;
            }

            position += 7;
        }

        Ok((value as i32, size))
    }
    /// Read VarLong as i64 with size in bytes (varint, size)
    fn read_varlong_size(&mut self) -> Result<(i64, usize), ProtocolError> {
        let mut value: u64 = 0;
        let mut position: u32 = 0;
        let mut size = 0;

        loop {
            let byte = self.read_byte()?;
            value |= ((byte & 0x7F) as u64) << position;

            size += 1;

            if (byte & 0x80) == 0 {
                break;
            }

            position += 7;
        }

        Ok((value as i64, size))
    }

    /// Read VarInt as i32
    fn read_varint(&mut self) -> Result<i32, ProtocolError> {
        self.read_varint_size().map(|o| o.0)
    }
    /// Read VarLong as i64
    fn read_varlong(&mut self) -> Result<i64, ProtocolError> {
        self.read_varlong_size().map(|o| o.0)
    }
}

impl<R: Read> DataReader for R {
    fn read_bytes(&mut self, size: usize) -> Result<Vec<u8>, ProtocolError> {
        let mut buf = vec![0; size];
        match self.read(&mut buf) {
            Ok(i) => if i == size {
                Ok(buf)
            } else if i == 0 {
                Err(ProtocolError::ConnectionClosedError)
            } else {
                buf.truncate(i);
                buf.append(&mut self.read_bytes(size-i)?);
                Ok(buf)
            },
            Err(_) => Err(ProtocolError::ReadError),
        }
    }
}
