use crate::ProtocolError;
use std::io::Write;
use uuid::Uuid;

/// Packet data writer trait
pub trait DataWriter {
    /// Write bytes
    fn write_bytes(&mut self, bytes: &[u8]) -> Result<(), ProtocolError>;

    /// Write byte
    fn write_byte(&mut self, byte: u8) -> Result<(), ProtocolError> {
        self.write_bytes(&[byte])
    }
    /// Write String
    fn write_string(&mut self, val: &str) -> Result<(), ProtocolError> {
        let bytes = val.as_bytes();
        self.write_varint(bytes.len() as i32)?;
        self.write_bytes(bytes)
    }
    /// Write UUID
    fn write_uuid(&mut self, val: &Uuid) -> Result<(), ProtocolError> {
        self.write_bytes(val.as_bytes())
    }
    /// Write Signed byte as i8
    fn write_signed_byte(&mut self, val: i8) -> Result<(), ProtocolError> {
        self.write_byte(val as u8)
    }
    /// Write Unsigned Short as u16
    fn write_unsigned_short(&mut self, val: u16) -> Result<(), ProtocolError> {
        self.write_bytes(&val.to_be_bytes())
    }
    /// Write Boolean
    fn write_boolean(&mut self, val: bool) -> Result<(), ProtocolError> {
        self.write_byte(if val { 0x01 } else { 0x00 })
    }
    /// Write Short as i16
    fn write_short(&mut self, val: i16) -> Result<(), ProtocolError> {
        self.write_bytes(&val.to_be_bytes())
    }
    /// Write Long as i64
    fn write_long(&mut self, val: i64) -> Result<(), ProtocolError> {
        self.write_bytes(&val.to_be_bytes())
    }
    /// Write Float as f32
    fn write_float(&mut self, val: f32) -> Result<(), ProtocolError> {
        self.write_bytes(&val.to_be_bytes())
    }
    /// Write Double as f64
    fn write_double(&mut self, val: f64) -> Result<(), ProtocolError> {
        self.write_bytes(&val.to_be_bytes())
    }
    /// Write Int as i32
    fn write_int(&mut self, val: i32) -> Result<(), ProtocolError> {
        self.write_bytes(&val.to_be_bytes())
    }

    /// Write VarInt as i32
    fn write_varint(&mut self, val: i32) -> Result<(), ProtocolError> {
        let mut value = val as u32;

        loop {
            if value & !0x7F == 0 {
                self.write_byte(value as u8)?;
                break;
            }
    
            self.write_byte(((value & 0x7F) | 0x80) as u8)?;
    
            value >>= 7;
        }

        Ok(())
    }
    /// Write VarLong as i64
    fn write_varlong(&mut self, val: i64) -> Result<(), ProtocolError> {
        let mut value = val as u64;

        loop {
            if value & !0x7F == 0 {
                self.write_byte(value as u8)?;
                break;
            }
    
            self.write_byte(((value & 0x7F) | 0x80) as u8)?;
    
            value >>= 7;
        }

        Ok(())
    }
}

impl<W: Write> DataWriter for W {
    fn write_bytes(&mut self, bytes: &[u8]) -> Result<(), ProtocolError> {
        self.write_all(bytes).map_err(|_| ProtocolError::WriteError)
    }
}